(***********************************************************************)
(*                                                                     *)
(*                              OCamlGen                               *)
(*                                                                     *)
(*                       OCaml module Generator                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 1997-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: walk_file.ml,v 1.13 2014/01/31 09:51:52 fclement Exp $ *)

let lexem_stream ic =
  let str = Lexing.from_channel ic in
  Stream.from (fun _ -> Some (Og_lexer.token str))
;;

let parse_channel ic =
  Og_parser.file (lexem_stream ic)
;;

let treat_channels treat (src, ic) (dst, oc) =
  try
    Og_debug.message "Parsing";
    let ast = parse_channel ic in
    Og_debug.message "Parsing done";
    if Og_debug.get_debug () then
      Og_types_print.file Format.err_formatter ast;
    Og_file.close_in src ic;
    Og_debug.message "Generating";
    treat ast;
    Og_debug.message "Generation done";
    Og_file.close_out dst oc;
    Og_file.make_read_only_file dst
  with
  | x ->
    Og_file.remove dst;
    raise x
;;

let get_channels (src, dst) =
  let src = Og_path.find src in
  let ic = Og_file.open_in src in
  Og_debug.message ("Input file " ^ Og_file.name_in src ^ " has been opened");
  let oc = Og_file.open_out dst in
  Og_debug.message ("Output file " ^ Og_file.name_out dst ^ " has been opened");
  (src, ic), (dst, oc)
;;

let parse_arguments suffix =
  let prog = Filename.basename Sys.executable_name in
  let usage_message =
    Printf.sprintf "\
      %s [-d | -keep] [-I include-dir] <source>\
    \n   or  [-o | -t] <destination> |\
    \n       [-c | -f] <source>"
      prog in

  Arg.parse [
    ("-I", Arg.String Og_path.push,
     "Add directory argument to the search path");
    ("-o", Arg.String Og_config.set_output_file_name, "Set output file name");
    ("-t", Arg.String Og_config.set_output_file_name, "Set output file name");
    ("-c", Arg.String Og_config.set_input_file_name, "Set input file name");
    ("-f", Arg.String Og_config.set_input_file_name, "Set input file name");
    ("-debug", Arg.Unit Og_debug.set_debug, "Fire debug mode");
    ("-d", Arg.Unit Og_debug.set_debug, "Fire debug mode");
    ("-i", Arg.Bool Og_config.set_ignore_explicit_alias,
           "Ignore explicit aliased type definitions (default is true).");
    ("-v", Arg.Unit Og_config.print_short_version,
     Printf.sprintf "Print the %s short version and exit" prog);
    ("-version", Arg.Unit Og_config.print_full_version,
     Printf.sprintf "Print the %s full version and exit" prog);
    ("-keep-type-definitions",
     Arg.Unit (fun () -> Og_config.set_keep_type_definitions true),
     "Keep type definitions in the target file\
    \n  (default is to remove type definitions).");
    ("-keep-comments",
     Arg.Unit (fun () -> Og_config.set_keep_comments true),
     "Keep comments in the target file\
    \n  (default is to remove comments).");
    ]
    (Og_config.set_anonymous_file_name suffix)
    usage_message;
  let files =
    try Og_config.get_input_output_file_names () with
    | Failure s ->
      prerr_endline s;
      prerr_endline "Usage:";
      prerr_endline usage_message;
      exit 2 in
  get_channels files
;;

module Main (P : Generator_spec.Prefix)
            (Treat : Walk_file_spec.Treat_walk) =

struct

  let (_src, _ic as input), (_dst, oc as output) =
    parse_arguments P.output_file_suffix
  ;;

  module Io : Walk_spec.Out =
    struct
      let buffer = Format.formatter_of_out_channel oc;;
    end
  ;;

  module Treatment = Treat (Io)
  ;;

  let main () =
    treat_channels Treatment.file input output;
    exit 0
  ;;

end
;;
