(***********************************************************************)
(*                                                                     *)
(*                              OCamlGen                               *)
(*                                                                     *)
(*                       OCaml module Generator                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 1997-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: og_types_print.ml,v 1.10 2014/01/31 09:51:52 fclement Exp $ *)

(* Og_types_print.ml: printing the data type definitions as found in the source
   file. *)

open Og_types;;
open Format;;

let comment s =
  let l = String.length s in
  if l >= 1 && s.[0] = '!'
  then String.sub s 1 (l - 1)
  else s
;;

let var_type ppf v = Format.fprintf ppf "'%s" v;;

let rec type_expr0 ppf = function
  | Var_type s -> fprintf ppf "(Var_type %a)" var_type s
  | Const_type ty_name ->
     fprintf ppf "(Const_type %a)" type_longident ty_name
  | Apply_type (_, _)
  | Tuple_type _
  | Function_type _ as ty -> fprintf ppf "(@[%a@])" (type_expr) ty
  | Paren_type ty -> fprintf ppf "(@[%a@])" (type_expr) ty

and type_application ppf = function
  | Apply_type (ty_name, ty_v) ->
    fprintf ppf "@[";
    begin match Array.length ty_v with
    | 0 -> ()
    | 1 -> type_application ppf ty_v.(0)
    | n ->
      fprintf ppf "(@[%a" (type_application) ty_v.(0);
      for i = 1 to n - 1 do
        fprintf ppf ",@ %a" (type_application) ty_v.(i)
      done;
      fprintf ppf "@])"
    end;
    fprintf ppf "@ %a@]" (type_longident) ty_name
  | Var_type _
  | Const_type _
  | Tuple_type _
  | Function_type _
  | Paren_type _ as ty ->
    type_expr0 ppf ty

and type_tuple ppf = function
  | Tuple_type tys ->
    fprintf ppf "@[<hv 2>(Tuple_type@ @[<hv 2>[|%a|]@])@]" (tuple_fields) tys
  | Var_type _
  | Const_type _
  | Apply_type _
  | Function_type _
  | Paren_type _ as ty ->
    type_application ppf ty

and type_function ppf = function
  | Function_type tys ->
    fprintf ppf "@[<hv 2>(Function_type@ (@[<hv 2>[|%a|]@]))@]"
       (function_fields) tys
  | Var_type _
  | Const_type _
  | Apply_type _
  | Tuple_type _
  | Paren_type _ as ty ->
    type_tuple ppf ty

and tuple_fields ppf fields =
  fprintf ppf "@[%a" (type_application) fields.(0);
  for i = 1 to Array.length fields - 1 do
    fprintf ppf ";@ %a" (type_application) fields.(i)
  done;
  fprintf ppf "@]"

and function_fields ppf fields =
  fprintf ppf "@[%a" (type_tuple) fields.(0);
  for i = 1 to Array.length fields - 1 do
    fprintf ppf ";@ %a" (type_tuple) fields.(i)
  done;
  fprintf ppf "@]"

and type_expr ppf = type_function ppf

and type_longident ppf = function
  | Lident id -> fprintf ppf "(Lident %s)" id
  | Ldot (longident, id) ->
    fprintf ppf "(Ldot (%a,%s))"
      (type_longident) longident id
  | Lapply (foncident, modident) ->
    fprintf ppf "(Lapply (%a, %a))"
      (type_longident) foncident (type_longident) modident
;;

let type_name ppf s = fprintf ppf "%s" s;;
let print_type_name = type_name;;
let print_type_expr = type_expr;;

let type_argument ppf s = fprintf ppf "(Var_type %s)" s;;

let type_arguments ppf = function
  | [] -> ()
  | [ s; ] ->
    fprintf ppf "%a " type_argument s
  | ss -> fprintf ppf "(%a) " (Lib_print.print_list ", " type_argument) ss
;;

let record_field ppf = function
  | Field (lab, ty_desc) ->
    fprintf ppf "@[<2>%s :@ %a@]" lab (type_expr) ty_desc
  | Field_mutable (lab, ty_desc) ->
    fprintf ppf "@[<2>mutable %s :@ %a@]" lab (type_expr) ty_desc
;;

let record_fields = Lib_print.print_list ";@ " record_field
;;

let sum_field ppf = function
  | Sum_tag lab ->
    fprintf ppf "| %s" lab
  | Sum_inj (lab, ty_desc) ->
    fprintf ppf "| @[<2>%s of@ %a@]" lab (type_expr) ty_desc
;;

let sum_fields = Lib_print.print_list "@ " sum_field
;;

let visibility ppf = function
  | Public -> fprintf ppf "%s@ " "Public"
  | Private -> fprintf ppf "%s@ " "Private"
  | Abstract -> fprintf ppf "%s@ " "Abstract"
;;

let print_visibility = visibility;;

let rec type_kind ppf = function
  | Record_type (vis, rfs) ->
    fprintf ppf
      "%a\
     %,  \
       @[<v>\
       {@[<v 1>@ \
         %a;\
        @]@ \
       }\
       @]"
      (visibility) vis
      (record_fields) rfs
  | Sum_type (vis, sfs) ->
    fprintf ppf
      "%a\
     %,  \
       @[<v>%a@]"
      (visibility) vis
      (sum_fields) sfs
  | Alias_type (vis, (ty, ty_desc_opt)) ->
    fprintf ppf
      "@[%a%a%a@]"
      (visibility) vis
      (type_expr) ty
      (Lib_print.print_option " =" type_kind) ty_desc_opt
  | External_type ->
    ()
;;

let type_def ppf (ty_name, ty_args, ty_kind) =
  fprintf ppf
    "@[<hov 2>([%a],@ %a,@ %a)@]"
    (type_arguments) ty_args
    (type_name) ty_name
    (type_kind) ty_kind
;;

let type_defs =
  Lib_print.print_list
    "@ \
     @ \
     and "
    type_def
;;

let dir ppf = function
  | Dir_open s -> fprintf ppf "Dir_open %s" s
  | Dir_include s -> fprintf ppf "Dir_include %s" s
;;

let phrase ppf = function
  | Comment s -> fprintf ppf "@[<hv 2>(Comment@ %s)@]@ " (comment s)
  | Directive d -> fprintf ppf "@[<hv 2>(Directive %a)@]@ " dir d
  | Type_definition ty_defs ->
    fprintf ppf
      "@[<v>\
       (Type_definition@;<1 2> %a)\
       @ \
       @ \
       @]\
      " (type_defs) ty_defs
;;

let file ppf phrases =
  fprintf ppf "@[<v>%a@]@."
    (fun ppf -> List.iter (phrase ppf))
    phrases
;;
