(* $Id: walk_input.ml,v 1.5 2012/06/18 00:01:15 weis Exp $ *)

open Format;;

(* An example of function to generate walker functions.
  The generator's concern is to generate pattern matchings and recursive
  calls to functions that walk the code. *)

(* Generation of input functions. *)

  let keywords = ref [];;
  let add_keyword s = keywords := s :: ! keywords;;

  let main_parser_name = ref "parser";;

  let quoted_escaped_string s = "\"" ^ String.escaped s ^ "\"";;

  let string_list ppf = function
  | [] -> ()
  | s :: sl ->
    fprintf ppf "%s" (quoted_escaped_string s);
    List.iter (fun s -> fprintf ppf ";@ %s" (quoted_escaped_string s)) sl;;

module Input : Action_spec.Make =
  functor (Io : Walk_spec.Out) ->

struct

  let ppf = Io.buffer
  ;;

  let print_seq () =
    Format.fprintf ppf ";@ "
  ;;

  let meta_input_string s =
    Format.fprintf ppf "@[<1>Scanf.bscanf ib@ \"%s\"@]" (String.escaped s)
  ;;

  let meta_input_char c =
    Format.fprintf ppf "Scanf.bscanf ib \"%c\" ()" c
  ;;

(*
  let quoted_string s = "\"" ^ s ^ "\""
  ;;
*)

  let add_sum_tag_rule _s = ();;
(*  let add_sum_inj_rule _s = ();; *)
  let sum_tag lab = add_keyword lab; add_sum_tag_rule lab
  let begin_sum_inj _lab = ();;
  let end_sum_inj _s = ();;

  let begin_tuple () =
    meta_input_char '('; print_seq ()
  ;;
    let begin_tuple_elem () = ()
    ;;
    let end_tuple_elem () = ()
    ;;
    let separator_tuple () =
      print_seq ();
      meta_input_char ',';
      print_seq ()
    ;;
  let end_tuple () =
    print_seq ();
    meta_input_char ')'
  ;;

  let begin_record_expr () =
    meta_input_char '{';
    print_seq ()
  ;;
  let end_record_expr () =
    meta_input_char '}'
  ;;

  let begin_record_field _lab =
    meta_input_string "\n "; print_seq ()
  ;;
  let end_record_field _lab =
    print_seq (); meta_input_char ';'; print_seq ()
  ;;

  let begin_application () =
    Format.fprintf ppf "@[<1>"
  ;;
    let separator_argument () = Format.fprintf ppf "@ "
    ;;
    let begin_argument () = ()
    ;;
    let end_argument () = ()
    ;;
  let end_application () =
    Format.fprintf ppf "@]"
  ;;

  let external_fun = function
    | "input_unit" -> Format.fprintf ppf "Lib_input.input_unit"
    | "input_bool" -> Format.fprintf ppf "Lib_input.input_bool"
    | "input_string" -> Format.fprintf ppf "Lib_input.input_quoted_string"
    | "input_int" -> Format.fprintf ppf "Lib_input.input_quoted_int"
    | "input_float" -> Format.fprintf ppf "Lib_input.input_quoted_float"
    | "input_char" -> Format.fprintf ppf "Lib_input.input_quoted_char"
    | "input_list" -> Format.fprintf ppf "Lib_input.input_List"
    | "input_array" -> Format.fprintf ppf "Lib_input.input_Array"
    | "input_option" -> Format.fprintf ppf "Lib_input.input_Option"
    | "input_fun" -> Format.fprintf ppf "Lib_input.input_fun"
    | "input_poly" -> Format.fprintf ppf "Lib_input.input_poly"
    | s -> Format.fprintf ppf "%s" s
  ;;

  let begin_paren () =
    meta_input_string "@[<1>(";
    print_seq ()
  ;;

  let end_paren () =
    print_seq ();
    meta_input_string "@,)@]"
  ;;

end
;;

module Gen : Walk_spec.Make =
  functor (Io : Walk_spec.Out) ->

struct

  let ppf = Io.buffer
  ;;

  let comment s =
    (* Comments have already been printer by the data type printer. *)
    if Config.get_keep_comments () &&
       not (Config.get_keep_type_definitions ())
    then fprintf ppf "%s@." (Og_types_pprint.comment s)
  ;;
  let directive s = fprintf ppf "open %s;;@.@." s;;

  let begin_file () =
    fprintf ppf
      "@[<v>(* Beginning of %s generated code *)@ @ "
      (Filename.basename Sys.executable_name);
    if not (Config.get_keep_type_definitions ())
    then directive (Config.get_input_module_name ())
  ;;

  let end_file () =
    let systeme_kwds =
      [ "("; ")"; "{"; "}"; "["; "]"; "[|"; "|]"; ","; ";";
        "true"; "false" ] in
    fprintf ppf
      "let lexer =@. make_lexer @[[%a]@];;@.@."
      (string_list) (systeme_kwds @ !keywords);
    fprintf ppf
       "\
       let lexem_stream ic =@.  \
         let str = Stream.of_channel ic in@.  \
         Stream.from (fun x -> Some (lexer str))@.\
       ;;@.@.\
       let parse_channel ic =@.  \
         %s (lexem_stream ic);;@.@.\
       let parse_string s =@.  \
         %s (lexer (Stream.of_string s))@."
      !main_parser_name !main_parser_name;

    fprintf ppf
      "(* End of %s generated code *)@]@."
      Sys.executable_name
  ;;

  let variable s = fprintf ppf "%s" s;;

  let begin_paren () = fprintf ppf "@[(";;
  let end_paren () = fprintf ppf "@,)@]";;

  let begin_application () = fprintf ppf "@[<1>";;
    let separator_argument () = fprintf ppf "@ ";;
    let begin_argument () = fprintf ppf "(";;
    let end_argument () = fprintf ppf ")";;
  let end_application () = fprintf ppf "@]";;

  let begin_tuple () = fprintf ppf "@[<1>(";;
    let separator_tuple () = fprintf ppf ",@ ";;
  let end_tuple () = fprintf ppf ")@]";;

  let begin_record_pattern () = fprintf ppf "@[<hv>{@[<hv>";;
    let begin_record_field f = fprintf ppf "@;<1 1>@[<2>%s =@ " f;;
    let end_record_field () = fprintf ppf "@]";;
    let separator_record_field () = fprintf ppf ";";;
  let end_record_pattern () = fprintf ppf "@]@ }@] ->@,";;

  let begin_def rec_flag =
    fprintf ppf "@[<v 0>le@[t %s"
      (if rec_flag = Walker_ast.Rec then "rec " else "")
  ;;
    let begin_def_elem ((_ty_name, walker_name), ty_arguments) =
      variable walker_name;
      List.iter (function arg -> fprintf ppf " %s" arg) ty_arguments;
    ;;
    let end_def_elem () = fprintf ppf "@]";;
    let separator_def_elem () = fprintf ppf "@,@,an@[d ";;
  let end_def () = fprintf ppf "@,;;@]@]@.@.";;

  let begin_function () = fprintf ppf "@[<1>(fun@ ";;
    let begin_clause_pattern () = fprintf ppf "@[";;
    let end_clause_pattern () = fprintf ppf "@] ->";;
    let begin_clause_expr () = fprintf ppf "@;<1 1>@[";;
    let end_clause_expr () =  fprintf ppf "@]";;
  let end_function () = fprintf ppf ")@]";;

  let sum_tag s = fprintf ppf "%s" s;;

  let begin_sum_inj s = add_keyword s; fprintf ppf "%s " s;;
  let end_sum_inj _s = ();;

  let begin_sum_clauses () =
    fprintf ppf " =@ function@;<1 -1>@[";;
    let begin_sum_clause () = fprintf ppf "@;<1 1>| @[";;
    let end_sum_clause () = fprintf ppf "@]";;
  let end_sum_clauses () = fprintf ppf "@]";;

  let begin_alias_clause () =
    fprintf ppf " =@ @[<1>";;
    let begin_aliased_clause () = fprintf ppf "@[<1>";;
    let end_aliased_clause () = fprintf ppf "@]";;
  let end_alias_clause () = fprintf ppf "@]";;

  let begin_record_clause () =
    fprintf ppf " =@ function@,@[<1>";;
  let end_record_clause () = fprintf ppf "@]";;

end
;;

(* Generates a value input program. *)
module Inputer =
 functor (Io : Walk_spec.Out) ->
 Walker.Make (Gen) (Input) (Io)
;;
